
-- # import prelude
-- # import id
-- # import nat
-- # import fin
-- # import vec
-- # import raw
-- # import type
-- # import typed


{--------------------------------------------------------------------------
    Type checking of raw terms
 --------------------------------------------------------------------------}

-- Checking terms ---------------------------------------------------------

data TypeError {n:Nat}(ctx:Ctx n) : RawTerm n -> Set where

  functionArgumentMismatch (s, s', t : Typ) :
      (e1 : Term ctx (s ==> t)) -> (e2 : Term ctx s') ->
      s =/= s' -> TypeError ctx (rapp (erase e1) (erase e2))

  notAFunction (e2 : RawTerm n) :
	       (e1 : Term ctx nat) -> TypeError ctx (rapp (erase e1) e2)

  errorInFunction (e1, e2 : RawTerm n) :
		  TypeError ctx e1 -> TypeError ctx (rapp e1 e2)

  errorInArgument (s, t : Typ)(e2 : RawTerm n) :
		  (e1 : Term ctx (s ==> t)) ->
		  TypeError ctx e2 -> TypeError ctx (rapp (erase e1) e2)

  errorInBody (s:Typ)(e : RawTerm (suc n)) :
	      TypeError (s::ctx) e -> TypeError ctx (rlam s e)

data TypeInference {n:Nat}(ctx:Ctx n) : RawTerm n -> Set where
  correct		    : (t:Typ) -> (e : Term ctx t)  -> TypeInference ctx (erase e)
  incorrect (e : RawTerm n) : TypeError ctx e -> TypeInference ctx e

infer (n:Nat) : (ctx:Ctx n) -> (e:RawTerm n) -> TypeInference ctx e
infer ctx (rvar i) = correct (ctx!i) (var i)

infer ctx (rapp e1 e2) with infer ctx e1, infer ctx e2 =>

      ctx (rapp _ _) (correct (s ==> t) e1) (correct s' e2) with compareType s s' =>
      
	  ctx (rapp (erase e1) (erase e2))
	      (correct (s ==> t) e1)
	      (correct s e2) same	    = correct t (e1 @ e2)

	  ctx (rapp (erase e1) (erase e2))
	      (correct (s ==> t) e1)
	      (correct s' e2) (different d) = incorrect (functionArgumentMismatch e1 e2 d)
      
      ctx (rapp (erase e1) e2) (correct (s ==> t) e1) (incorrect err)
					    = incorrect (errorInArgument e1 err)

      ctx (rapp (erase e1) _) (correct nat e1) _
					    = incorrect (notAFunction e1)

      ctx (rapp e1 _) (incorrect err) _	    = incorrect (errorInFunction err)

infer ctx (rlam s e) with infer (s::ctx) e =>
      ctx (rlam s (erase e)) (correct t e)    = correct (s ==> t) (lam e)
      ctx (rlam s  e       ) (incorrect err)  = incorrect (errorInBody err)


{--------------------------------------------------------------------------
    Testing
 --------------------------------------------------------------------------}

rK (n:Nat) : Typ -> Typ -> RawTerm n
rK s t = rlam s $ rlam t $ rvar (fsuc fzero)

