\documentclass{article}

\usepackage{haskell}

\begin{document}

\begin{code}
-- # import Property

data List (A : Set) : Set where
  nil : List A
  (::) : A -> List A -> List A
infixr 10 ::

-- Values
Int : Set

infixl 20 +
(+) : Int -> Int -> Int

data V : Set where
  int : Int -> V

propInjInt : (n, m : Int) -> Eq V (int n) (int m) -> Eq Int n m

-- Stack
S = List V

-- Security levels
data Sec : Set where
  hi  : Sec
  lo  : Sec

lub : Sec -> Sec -> Sec
lub hi s  = hi
lub s hi  = hi
lub lo lo = lo

-- Type
data T : Set where
  tInt : Sec -> T

-- Stack type
ST = List T

-- Environment
E = S

-- Environment types
ET = ST

-- Instructions
data I : Set where
  add : I

-- Evaluation
(-->) : {S;I} -> S -> Prop
(int v1 :: int v2 :: s, add) --> (int v :: s') = Eq Int v (v1 + v2) /\ Eq S s s'
_ --> _					    = False

-- Low equivalence
lowEqV : T -> V -> V -> Prop
lowEqV (tInt hi) v w = True
lowEqV (tInt lo) v w = Eq V v w

lowEqS : ST -> S -> S -> Prop
lowEqS [] [] []		      = True
lowEqS (t::st) (v1::s1) (v2::s2) = lowEqV t v1 v2 /\ lowEqS st s1 s2
lowEqS [] [] (_::_)	      = False
lowEqS [] (_::_) []	      = False
lowEqS (_::_) [] []	      = False
lowEqS [] (_::_) (_::_)	      = False
lowEqS (_::_) [] (_::_)	      = False
lowEqS (_::_) (_::_) []	      = False

lowEqE = lowEqS

-- Typing relation
TypeCorrect : {ET;I} -> ET -> Prop
TypeCorrect (tInt s1 :: tInt s2 :: st, add) (tInt s :: st') =
    Eq Sec s (lub s1 s2) /\ Eq ST st st'
TypeCorrect _ _ = False

typeInversion :
    (g : ET) -> (i : I) -> (g' : ET) ->
    (P : ET -> I -> ET -> Prop)	->
    TypeCorrect (g,i) g'		->
    ((s1,s2 : Sec) -> (st : ST) -> P	(tInt s1 :: tInt s2 :: st)
					add
					(tInt (lub s1 s2) :: st)
    ) -> P g i g'
typeInversion (tInt s1 :: tInt s2 :: st)
	      add
	      (tInt s :: st')
	      P typeder
	      branch = fol-plugin(typeder, branch)
typeInversion g i g' P typeder branch = fol-plugin(typeder)

typeInversion' :
    (g : ET) -> (i : I) -> (g' : ET) ->
    (P : ET -> I -> ET -> Type)	->
    TypeCorrect (g,i) g'		->
    ((s1,s2 : Sec) -> (st : ST) -> P	(tInt s1 :: tInt s2 :: st)
					add
					(tInt (lub s1 s2) :: st)
    ) -> P g i g'

{-
data TypeCorrect (g : ET)(i : I)(g' : ET) : Prop where
  typeAdd :  (s1, s2 : Sec) -> (g : ET) ->
	      TypeCorrect (tInt s1 :: tInt s2 :: g) add
			  (tInt (lub s1 s2) :: g)
-}

-- Non interference
NI : I -> ET -> ET -> Type
NI i g g' =
  (e1, e2 : E) -> lowEqE g e1 e2 ->
  { e1', e2' : E
  ; lowEqE g' e1' e2' /\
    ( (e1,i) --> e1' <=> (e2,i) --> e2' )
  }

-- Lemma
lemma :
  (s, s' : Sec) -> (n, n', m, m' : Int) ->
  (f : Int -> Int -> Int) ->
  lowEqV (tInt s) (int n) (int m) ->
  lowEqV (tInt s') (int n') (int m') ->
  lowEqV (tInt (lub s s')) (int (f n n')) (int (f m m'))
lemma hi s' = fol-plugin()
lemma s hi  = fol-plugin()
lemma lo lo = fol-plugin(propInjInt)

-- Da Property

theProof :
    (g : ET) -> (i : I) -> (g' : ET) ->
    TypeCorrect (g, i) g' -> NI i g g'
theProof g i g' typeder =
  typeInversion' g i g' (\g0 i0 g0' -> NI i0 g0 g0')
      typeder
      (\s1 s2 st -> ?)
{-
(tInt s1 :: tInt s2 :: g)
	  add
	  (int v1 :: int v1' :: e1)
	  (int v2 :: int v2' :: e2) =
    \loweq -> ( int (v1 + v1') :: e1
	      , int (v2 + v2') :: e2
	      , (fol-plugin(lemma, loweq), fol-plugin())
	      )
-}
\end{code}

\end{document}
